#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['b8']

package_data = \
{'': ['*']}

install_requires = \
['msgpack >= 1.0.0']

entry_points = \
{'console_scripts': ['b8 = b8.app:main']}

setup(name='b8',
      version='0.1.10',
      description='b8, Bominade Ultralight NeoVim-based IDE',
      author='Ali Afshar',
      author_email='aafshar@gmail.com',
      url='https://gitlab.com/afshar-oss/b8',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.5',
     )
